from typing import Any,Iterable,Union
#from collections import namedtuple
import yaml
import os
import os.path
import math
import numpy as np
import sasktran as sk
from skimpy import FOV_Rectangle,Wavelength_GaussianPSF, Pixel,Instrument, ForwardModel
from skimpy.platform.platform_onepoint_nadir import Platform_OnePoint_Nadir
from .detector_hyvisi_chroma import DetectorProperties_Hyvisi_Chroma
from .aimnorth_l1generator import AimNorth_Level1, encode_aim_north_level1_data



#------------------------------------------------------------------------------
#           read_yaml_file
#------------------------------------------------------------------------------

def read_yaml_file(fname: str):
    fullname = __file__                                         # Get the filename of this module
    [basedir, spare] = os.path.split(fullname)                  # split out the list_of_files
    fullname = basedir + os.sep + 'data' + os.sep + fname
    try:
        f = open(fullname, 'rt')
        configuration = yaml.load(f)
        f.close()
    except Exception as inst:
        print(type(inst))  # the exception instance
        print(inst.args)  # arguments stored in .args
        print(inst)  # __str__ allows args to be printed directly,
        raise RuntimeError(
            'Error reading the YAML configuration file %s from file %s' % (fname, fullname,))
    return configuration


#------------------------------------------------------------------------------
#           AimNorth_Simulator
#------------------------------------------------------------------------------

class AimNorth_Simulator(object):
    """
    The following settinsg are provided

        * aperture_diameter_cms
        * wavelengths
        * psf
        * pixel_iyfov
        * pixel_izfov
        * z_icf_bearing
        * numhighreslos
        * numhighreswave

    """

    #------------------------------------------------------------------------------
    #           __init__
    #------------------------------------------------------------------------------

    def __init__(self):
        """
        Creates the Aim-North Simulator object.
        """
        self.instrument_settings = self._load_default_instrument_settings()
        self.platform            =  Platform_OnePoint_Nadir()
        self.detector            =  DetectorProperties_Hyvisi_Chroma()

    #------------------------------------------------------------------------------
    #           set_pixel_center_wavelengths
    #------------------------------------------------------------------------------

    def set_pixel_center_wavelengths(self, wavelengths: Union[np.ndarray, Iterable] ):
        """
        Sets the center wavelengths of the spectrograph pixels. There are two options
        (i) provide an array of wavelengths at whatever points the user chooses or
        (ii) provide an array of 3 elements suitable for numpy.arange (start,stop, stepsize).
        The wavelengths are stored in `instrument_settings[wavelengths]`

        Parameters
        ----------
        wavelengths: np.ndarray
            An array specifying the central wavelength of each pixel used in the spectrograph simulation. It can either be
            just array of wavelengths (nm) or it can be a 3 element array compatible with numpy.arange, in which case we
            check the last element to make sure its less than 10 otehrwise we assume its an array of 3 wavelength points

        """
        w = np.array(wavelengths)
        if (w.size ==3) and (w[2] < 10.0):
            w = np.arange( w[0],w[1],w[2])
        self.instrument_settings['wavelengths'] = w

    #------------------------------------------------------------------------------
    #           set_instrument_attribute
    #------------------------------------------------------------------------------

    def set_instrument_attribute(self, key:str, value:Any):

        if key == 'wavelengths':
            self.set_pixel_center_wavelengths(value)
        else:
            oldval = self.instrument_settings.get(key)
            if (oldval is None):
                raise ValueError("You can only set attributes of existing keys. Key, %s, does not exist"%(key,) )
            self.instrument_settings[key] = value

    #------------------------------------------------------------------------------
    #           _load_default_instrument_settings
    #------------------------------------------------------------------------------

    def _load_default_instrument_settings(self):

        self.instrument_settings = {}
        config = read_yaml_file('aim_north_settings.yaml')                                               # read aim-north configuration
        for k,value in config.items():
            self.instrument_settings[k] = value
        w = self.instrument_settings['wavelengths']
        self.set_pixel_center_wavelengths(w)
        return self.instrument_settings

     #------------------------------------------------------------------------------
    #           make_forwardmodel
    #------------------------------------------------------------------------------

    def _make_forwardmodel( self, instrument : Instrument  )->ForwardModel:
        """
        Makes and configures the sasktran forward model for the Aim North simulations. This typically means configuring
        the atmosphere used for the Sasktran Radiative transfer calculation and setting up the call back function so it
        calls our method to make Level 1 products once radiancesetc. have been calculates.
        """

        atmosphere          = sk.Atmosphere()
        atmosphere['air']   = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmosphere['ozone'] = sk.Species(sk.O3DBM(), sk.Labow())

        model = ForwardModel( level1_supplier_func = encode_aim_north_level1_data)
        model.radiance_simulator().set_atmosphere( atmosphere )
        model.radiance_simulator().set_option_weighting_functions_for_toa_albedo( True )
        return model

    #------------------------------------------------------------------------------
    #           make_AimNorth_instrument
    #------------------------------------------------------------------------------

    def _make_instrument( self ):                                                                                        # The number of high resolution wavelengths across the point spread function of each wavelength pixel

        """
        Makes the instrument for Aim-North. This is primarily a matter of defining the field of views of the pixels
        and the wavelength bandpass for various pixels.
        """

        instrument            = Instrument()                                                                            # Create the skimpy instrument object
        aperture_diameter_cms = self.instrument_settings['aperture_diameter_cms']
        wavelengths           = self.instrument_settings['wavelengths']
        psf                   = self.instrument_settings['psf']
        pixel_iyfov           = self.instrument_settings['pixel_iyfov']
        pixel_izfov           = self.instrument_settings['pixel_izfov']
        numhighreslos         = self.instrument_settings['numhighreslos']
        numhighreswave        = self.instrument_settings['numhighreswave']


        fer = instrument.front_end_radiance()
        fer.set_aperture_area_cm2( aperture_diameter_cms*aperture_diameter_cms*math.pi/4.0 )                            # Set up the entrace aperture area in cm2

        fov = FOV_Rectangle( Hrange=(numhighreslos, 0.0, pixel_iyfov),
                             Vrange=(numhighreslos, 0.0, pixel_izfov))                                                   # Define AimNorth AQ field of view of 1 pixel. Use just one line of sight for horizontal and vertical
        instrument.front_end_radiance().add_field_of_view( fov)

        wav = list(wavelengths)                                                                                         # Convert the incoming wavelengths into a list
        for w in wav:                                                                                                   # and
            fer.add_wavelength_range( Wavelength_GaussianPSF(numhighreswave, w, psf))                                   # add the wavelength range to our frontendradiance object

        numpix = len(wav)                                                                                               # define all the "spectrum" pixels, one pixel for each wavelength
        for i in range(numpix):                                                                                         # so for each wavelength
            fer.add_pixel( Pixel( fer.fov[0], fer.wavelength_ranges[i]))                                                # create the pixels using the same field of view and different wavelength ranges

        return instrument

    #------------------------------------------------------------------------------
    #           set_platform_location
    #------------------------------------------------------------------------------

    def set_platform_location(self, latitude:float, longitude:float, utc: np.datetime64,  height_meters:float=37770000.0 ):
        """
        Sets the time and location of the platform using latitude, longitude, time and optional height. Ensures the
        platform is running in nadir mode.

        Parameters
        ----------
        latitude : float
            The latitude of the platform in degrees

        longitude : float
            The longitude of the platform in degrees

        utc : numpy.datetime64
            The UTC time of the measurements.

        height_meters : float
            Optional, default is 37770000.0. The height of the satellite in meters above the surface.

        """
        self.platform.set_platform_location( latlonheightandt= (latitude,longitude,height_meters, utc) )
        self.platform.set_nadir_mode(True)

    #------------------------------------------------------------------------------
    #           set_ground_targets
    #------------------------------------------------------------------------------

    def set_ground_target_locations(self, target_latitudes: np.ndarray, target_longitudes: np.ndarray, clear_states : bool = True ):

        """
        :param target_latitudes:
        :param target_longitudes:
        :param clear_states:
        :return:
        """
        zazi = self.instrument_settings['z_icf_bearing']
        if (clear_states): self.platform.clear_states()
        for i in range(target_latitudes.size):
            self.platform.set_boresight_to_look_at_geodetic_location( target_latitudes[i],
                                                                      target_longitudes[i],
                                                                      0.0,
                                                                      nadir_azimuth_of_inst_z = zazi)
            self.platform.add_current_platform_state()

    #------------------------------------------------------------------------------
    #           calculate_radiance
    #------------------------------------------------------------------------------

    def calculate_radiance(self)->AimNorth_Level1:

        instrument   = self._make_instrument()
        forwardmodel = self._make_forwardmodel( instrument )
        #forwardmodel.radiance_simulator().set_option_generate_dummy_data(True)

        level1 = forwardmodel.calculate_radiance( instrument, self.platform)
        return level1

